#![no_std]
#![no_main]
#![feature(custom_test_frameworks)]
#![test_runner(rustos::test_runner)]
#![reexport_test_harness_main = "test_main"]

use bootloader::{entry_point, BootInfo};
use core::panic::PanicInfo;
use rustos::println;

// define entry point for bootloader
entry_point!(kernel_main);

// fn kernel_main :::
// boot_info: &'static BootInfo :param:
fn kernel_main(boot_info: &'static BootInfo) -> ! {
    use rustos::memory;
    use x86_64::structures::paging::Translate;
    use x86_64::VirtAddr;

    println!("Hello World{}", "!");
    rustos::init(); // initializes kernel

    let phys_mem_offset = VirtAddr::new(boot_info.physical_memory_offset);

    let mapper = unsafe { memory::init(phys_mem_offset) };

    let addresses = [
        // identity-mapped vga buffer
        0xb8000,
        // some code page
        0x201008,
        // some stack page
        0x0100_0020_1a10,
        // VirtAddr mapped to PhysAddr 0
        boot_info.physical_memory_offset,
    ];

    for &address in &addresses {
        let virt = VirtAddr::new(address);
        let phys = mapper.translate_addr(virt);
        println!("{:?} -> {:?}", virt, phys);
    }

    #[cfg(test)]
    test_main();

    println!("Hey, it didn't crash!");
    rustos::hlt_loop();
}

// fn panic ::: prints panic message
// info: &PanicInfo :param: passed by panic_handler, includes panic information
#[panic_handler]
fn panic(info: &PanicInfo) -> ! {
    #[cfg(test)]
    rustos::test_panic_handler(info);
    #[cfg(not(test))]
    {
        println!("{}", info);
        rustos::hlt_loop();
    };
}
