use lazy_static::lazy_static;
use spin::Mutex;
use uart_16550::SerialPort;

// macro serial_print ::: prints arg to serial port
#[macro_export]
macro_rules! serial_print {
  ($($arg:tt)*) => { $crate::serial::_print(format_args!($($arg)*)); };
}

// macro serial_println ::: prints arg to serial port with \n appended
#[macro_export]
macro_rules! serial_println {
  () => ($crate::serial_print!("\n"));
  ($fmt:expr) => ($crate::serial_print!(concat!($fmt, "\n")));
  ($fmt:expr, $($arg:tt)*) => ($crate::serial_print!( concat!($fmt, "\n"), $($arg)*));
}

lazy_static! {
    // SERIAL1: Mutex<SerialPort> ::: serial port to write to
    pub static ref SERIAL1: Mutex<SerialPort> = {
        // SAFETY ensure correct SerialPort is initialized to write output to
        let mut serial_port = unsafe { SerialPort::new(0x3F8) };
        serial_port.init();
        Mutex::new(serial_port)
    };
}

// fn _print ::: prints to serial port; internal fn
// args: ::core::fmt::Arguments :param: goofy ah macro magic
#[doc(hidden)]
pub fn _print(args: ::core::fmt::Arguments) {
    use core::fmt::Write;
    use x86_64::instructions::interrupts;

    // locked to avoid race conditions
    interrupts::without_interrupts(|| {
        SERIAL1
            .lock()
            .write_fmt(args)
            .expect("Printing to serial failed");
    });
}
